import os
import re
from datetime import date, datetime
from pathlib import Path, PosixPath
from typing import Iterable, List

import numpy as np
from astropy.io.fits import getdata as getfitsdata
from numpy.typing import DTypeLike

from m23.constants import (
    INPUT_NIGHT_FOLDER_NAME_DATE_FORMAT,
    OUTPUT_NIGHT_FOLDER_NAME_DATE_FORMAT,
)

# local imports
from .rename import rename


def get_image_number_in_log_file_combined_file(file: Path) -> int:
    """
    Returns the image number log file combined file, or raises error if image number not found
    Examples:
        In the filename, 07-07-18_m23_7.0-112.txt, the image number is 112
    """
    results = re.findall(r"^.*-(\d+)\.txt", file.name)
    if len(results) == 0:
        raise ValueError(f"{file.name} is not in something-xxx.txt format")
    else:
        return int(results[0])


def get_image_number_in_fit_file(file: Path) -> int:
    """
    Returns the image number of the fit file, or raises error if image number not found
    Examples:
        In the filename, m23_7.0-010.fit, image number is 10
        More generally, In something-xxx.fit, integer representing xxx defines the image number
    """
    results = re.findall(r"^.*-(\d+)\.fit", file.name)
    if len(results) == 0:
        raise ValueError(f"{file.name} is not in something-xxx.fit format")
    else:
        return int(results[0])


def get_flats(folder: Path) -> Iterable[PosixPath]:
    """
    Return a list of flat files in `folder` provided
    """
    return folder.glob("*flat*.fit")


def get_darks(folder: Path) -> Iterable[PosixPath]:
    """
    Return a list of dark files in `folder` provided
    """
    return folder.glob("*dark*.fit")


def get_all_fit_files(folder: Path) -> Iterable[PosixPath]:
    """
    Return a list of all fit files in `folder` provided
    """
    return folder.glob("*.fit")


def get_raw_images(folder: Path) -> Iterable[PosixPath]:
    """
    Return a list of all fit files sorted by asc. by image number present in filename
    For what defines the image number, see the definition of `get_image_number_in_fit_file`
    function. Note that files without digit are displayed at the top
    """
    return sorted(folder.glob("*.fit"), key=get_image_number_in_fit_file)


def get_radius_folder_name(radius: int) -> str:
    """
    Returns the folder name to use for a given radius pixel of extraction
    """
    radii = {
        1: "One",
        2: "Two",
        3: "Three",
        4: "Four",
        5: "Five",
        6: "Six",
        7: "Seven",
        8: "Eight",
        9: "Nine",
    }
    if result := radii.get(radius):
        return f"{result} Pixel Radius"
    else:
        return f"{radius} Pixel Radius"


def get_date_from_input_night_folder_name(name: str | Path) -> date:
    if issubclass(type(name), Path):
        name = name.name
    return datetime.strptime(name, INPUT_NIGHT_FOLDER_NAME_DATE_FORMAT).date()


def get_output_folder_name_from_night_date(night_date: date) -> str:
    return night_date.strftime(OUTPUT_NIGHT_FOLDER_NAME_DATE_FORMAT)


def fitFilesInFolder(folder, fileType="All"):
    allFiles = os.listdir(folder)
    fileType = fileType.lower()

    allFiles = list(filter(lambda x: x.endswith(".fit"), allFiles))
    if fileType == "all":
        return allFiles
    else:
        return list(filter(lambda x: x.__contains__(fileType), allFiles))


def fitDataFromFitImages(images):
    return [getfitsdata(item) for item in images]


def fit_data_from_fit_images(images: Iterable[str | Path]) -> List[DTypeLike]:
    return [getfitsdata(item) for item in images]


def get_log_file_name(night_date: date):
    return f"Night-{night_date}-Processing-log.txt"


### Similar to the default version of IDL Median
### https://github.com/LutherAstrophysics/python-helpers/issues/8
def customMedian(arr, *args, **kwargs):
    arr = np.array(arr)
    if len(arr) % 2 == 0:
        newArray = np.append(arr, [np.multiply(np.ones(arr[0].shape), np.max(arr))], axis=0)
        return np.median(newArray, *args, **kwargs)
    else:
        return np.median(arr, *args, **kwargs)


__all__ = [
    "customMedian",
    "fitFilesInFolder",
    "rename",
    "get_closet_date",
    "raw_data_name_format",
]



### surroundWithZeros
###
### parameters:
###   npArray: numpy array
###   size: integer specifying the length of the edge
###   value: (optional) value to surround with, 0 default
###
###
### returns: None
###
### it mutates the original array provided, npArray
def surroundWith(npArray, size, value=0):

    # columns
    npArray[:, :size] = value
    npArray[:, -size:] = value

    # rows
    npArray[:size, :] = value
    npArray[-size:, :] = value



###  Taken from https://stackoverflow.com/questions/44782476/split-a-numpy-array-both-horizontally-and-vertically
###
### arr is the numpy array to turn into blocks
### blockSize is the tuple/array of sizes
def blockRegions(arr, blockSize):
    # arr is input array, blockSizeis block-size
    m, n = arr.shape
    M, N = blockSize
    return arr.reshape(m // M, M, n // N, N).swapaxes(1, 2).reshape(-1, M, N)



from .crop import crop, cropIntoRectangle
from .fill import fillMatrix
from .region import blockRegions
from .utils import surroundWith

__all__ = ["crop", "cropIntoRectangle", "fillMatrix", "surroundWith", "blockRegions"]



from pathlib import Path
from typing import List


def internight_normalize(
    night: Path, reference_file: Path, color_file: Path, radii_of_extraction: List[int]
) -> None:
    """
    This function normalizes the Flux Logs Combined for a night with respect to
    the data in the reference night. It also saves the result of inter-night normalization.
    We typically the image 71 on Aug 4 2003 night as the reference.

    Note that since the a night can have Flux Logs Combined for multiple radii of extraction,
    this functions creates a color normalized output folder for as many radii of extraction
    as specified. Note that for each specified radius of extraction specified, respective
    Flux Logs Combined folder has to exist in `night` Path provided.

    param: night: Night folder that contains Flux Logs Combined folder
    param: reference_file: The path to the reference file to use.
    param: color_file: The path to the the mean R-I file to use.

    return: None

    Side effects:

    This function creates a 'Color Normalized' folder inside `night` Path and
    folders for each radius specified in radii_of_extraction inside 'Color Normalized' folder
    that contains a txt file of the inter-night normalized data for each star for that night.

    This function also logs to the default log file in `night`. See `process_nights` inside `processor`
    module for the details of the log file.


    Preconditions:
    `night` is a valid path containing Flux Logs Combined for all radius specified in `radii_of_extraction`
    `reference_file` is a valid file path in conventional reference file format
    `color_file` is a valid file path in conventional R-I color file format
    """

    for radius in radii_of_extraction:
        internight_normalize_auxiliary(night, reference_file, color_file, radius)


def internight_normalize_auxiliary(
    night: Path, reference_file: Path, color_file: Path, radius_of_extraction: List[int]
):
    """
    This is an auxiliary function for internight_normalize that's different from the
    `internight_normalize` because this function takes `radius_of_extraction` unlike
    `internight_normalize` that takes `radii_of_extraction`.

    See https://drive.google.com/file/d/1R1Xc9RhPEYXgF5jlmHvtmDqvrVWs6xfK/view?usp=sharing
    for explanation of this algorithm by Prof. Wilkerson.
    """
    pass



from datetime import date
from pathlib import Path


class MasterflatFile:
    @classmethod
    def generate_file_name(cls, night_date: date) -> str:
        """
        Generates a masterflat filename based on the night date

        Param: night_date: The date of the night for which to generate filename
        Returns: Filename string
        """
        date_string = night_date.strftime("%Y-%m-%d")  # YYYY-MM-DD format
        return f"{date_string}_masterflat.fit"

    def __init__(self, file_path: Path) -> None:
        self.__path = file_path
        self.__data = None

    def _validate_file(self):
        """Validates that the file exists"""
        if not self.path().exists():
            raise FileNotFoundError(f"{self.path()}")
        if not self.path().is_file():
            ValueError(f"{self.path()} isn't a file")
        if not self.path().suffix == ".fit":
            ValueError(f"{self.path()} isn't a fit file")

    def _read_data(self):
        # TODO
        pass

    def path(self):
        return self.__path

    def data(self):
        self._validate_file()
        return self.__data

    def save_file(self, data, header=None):
        # TODO
        pass

    def get_data(self, data, header=None):
        # TODO
        pass



from .generate_masterflat import generate_masterflat
from .process_nights import start_data_processing
from .renormalize import renormalize

__all__ = ["start_data_processing", "renormalize", "generate_masterflat"]



import logging
import sys
from datetime import date
from pathlib import Path
from typing import List

import toml
from astropy.io.fits import getdata

from m23.align.alignment import imageAlignment
from m23.calibrate.calibration import calibrateImages
from m23.calibrate.master_calibrate import makeMasterDark
from m23.combine.combination import imageCombination
from m23.constants import (
    ALIGNED_COMBINED_FOLDER_NAME,
    CONFIG_FILE_NAME,
    FLUX_LOGS_COMBINED_FOLDER_NAME,
    INPUT_CALIBRATION_FOLDER_NAME,
    LOG_FILE_COMBINED_FILENAME_DATE_FORMAT,
    LOG_FILES_COMBINED_FOLDER_NAME,
    M23_RAW_IMAGES_FOLDER_NAME,
    MASTER_DARK_NAME,
    MASTER_FLAT_NAME,
    OUTPUT_CALIBRATION_FOLDER_NAME,
)
from m23.extract.extraction import extract_stars
from m23.matrix import crop
from m23.matrix.fill import fillMatrix
from m23.norm.normalize import normalizeLogFiles
from m23.processor.config_loader import Config, ConfigInputNight, validate_file
from m23.utils import (
    fit_data_from_fit_images,
    get_darks,
    get_date_from_input_night_folder_name,
    get_flats,
    get_log_file_name,
    get_output_folder_name_from_night_date,
    get_radius_folder_name,
    get_raw_images,
)


def normalization_helper(
    radii_of_extraction: List[int],
    FLUX_LOGS_COMBINED_OUTPUT_FOLDER: Path,
    ref_file_path: Path,
    log_files_to_use: List[Path],
    night_date: date,
):
    """
    This is a normalization helper function extracted so that it can be reused by the renormalization script
    """
    for index, radius in enumerate(radii_of_extraction):
        logging.info(f"Normalizing for radius of extraction {radius} px")
        RADIUS_FOLDER = FLUX_LOGS_COMBINED_OUTPUT_FOLDER / get_radius_folder_name(radius)
        RADIUS_FOLDER.mkdir(exist_ok=True)  # Create folder if it doesn't exist
        for file in RADIUS_FOLDER.glob("*"):
            if file.is_file():
                file.unlink()  # Remove each file in the folder
        logfile_adu_column = 6 + index  # 6th column is the first column with Star ADU
        normalizeLogFiles(
            ref_file_path,
            log_files_to_use,
            logfile_adu_column,
            RADIUS_FOLDER,
            night_date,
        )


def process_night(night: ConfigInputNight, config: Config, output: Path, night_date: date):
    """
    Processes a given night of data based on the settings provided in `config` dict
    """
    # Save the config file used to do the current data processing
    CONFIG_PATH = output / CONFIG_FILE_NAME
    with CONFIG_PATH.open("w+") as fd:
        toml.dump(config, fd)

    # Number of expected rows and columns in all raw images
    rows, cols = config["image"]["rows"], config["image"]["columns"]
    radii_of_extraction = config["processing"]["radii_of_extraction"]

    log_file_path = output / get_log_file_name(night_date)
    # Clear file contents if exists, so that reprocessing a night wipes out contents instead of appending to it
    if log_file_path.exists():
        log_file_path.unlink()
    logging.basicConfig(
        filename=log_file_path,
        format="%(asctime)s %(message)s",
        level=logging.INFO,
    )
    # Write to std out in addition to writing to a logfile
    logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))
    logging.info(f"Starting processing for {night_date}")

    ref_image_path = config["reference"]["image"]
    ref_file_path = config["reference"]["file"]

    # Define relevant input folders for the night being processed
    NIGHT_INPUT_FOLDER: Path = night["path"]
    NIGHT_INPUT_CALIBRATION_FOLDER: Path = NIGHT_INPUT_FOLDER / INPUT_CALIBRATION_FOLDER_NAME
    NIGHT_INPUT_IMAGES_FOLDER = NIGHT_INPUT_FOLDER / M23_RAW_IMAGES_FOLDER_NAME

    # Define and create relevant output folders for the night being processed
    CALIBRATION_OUTPUT_FOLDER = output / OUTPUT_CALIBRATION_FOLDER_NAME
    ALIGNED_COMBINED_OUTPUT_FOLDER = output / ALIGNED_COMBINED_FOLDER_NAME
    LOG_FILES_COMBINED_OUTPUT_FOLDER = output / LOG_FILES_COMBINED_FOLDER_NAME
    FLUX_LOGS_COMBINED_OUTPUT_FOLDER = output / FLUX_LOGS_COMBINED_FOLDER_NAME

    for folder in [
        CALIBRATION_OUTPUT_FOLDER,
        ALIGNED_COMBINED_OUTPUT_FOLDER,
        LOG_FILES_COMBINED_OUTPUT_FOLDER,
        FLUX_LOGS_COMBINED_OUTPUT_FOLDER,
    ]:
        if folder.exists():
            [file.unlink() for file in folder.glob("*") if file.is_file()]  # Remove existing files
        folder.mkdir(exist_ok=True)

    crop_region = config["image"]["crop_region"]

    # Darks
    darks = fit_data_from_fit_images(get_darks(NIGHT_INPUT_CALIBRATION_FOLDER))
    if len(crop_region) > 0:
        darks = [crop(matrix, rows, cols) for matrix in darks]
    master_dark_data = makeMasterDark(
        saveAs=CALIBRATION_OUTPUT_FOLDER / MASTER_DARK_NAME,
        headerToCopyFromName=next(get_darks(NIGHT_INPUT_CALIBRATION_FOLDER)).absolute(),
        listOfDarkData=darks,
    )
    logging.info(f"Created master dark")
    del darks  # Deleting to free memory as we don't use darks anymore

    # Flats
    if night.get("masterflat"):
        master_flat_data = getdata(night["masterflat"])
        logging.info(f"Using pre-provided masterflat")
    else:
        flats = fit_data_from_fit_images(get_flats(NIGHT_INPUT_CALIBRATION_FOLDER))
        if len(crop_region) > 0:
            flats = [crop(matrix, rows, cols) for matrix in flats]

        master_flat_data = makeMasterDark(
            saveAs=CALIBRATION_OUTPUT_FOLDER / MASTER_FLAT_NAME,
            headerToCopyFromName=next(
                get_flats(NIGHT_INPUT_CALIBRATION_FOLDER)
            ).absolute(),  # Gets absolute path of first flat file
            listOfDarkData=flats,
        )
        logging.info(f"Created masterflat")
        del flats  # Deleting to free memory as we don't use flats anymore

    raw_images: List[Path] = list(get_raw_images(NIGHT_INPUT_IMAGES_FOLDER))
    logging.info(f"Processing images")
    no_of_images_to_combine = config["processing"]["no_of_images_to_combine"]
    logging.info(f"Using no of images to combine: {no_of_images_to_combine}")
    logging.info(f"Radii of extraction: {radii_of_extraction}")

    # We now Calibrate/Crop/Align/Combine/Extract set of images in the size of no of combination
    # Note the subtle typing difference between no_of_combined_images and no_of_images_to_combine
    no_of_combined_images = len(raw_images) // no_of_images_to_combine

    for i in range(no_of_combined_images):
        from_index = i * no_of_images_to_combine
        to_index = (i + 1) * no_of_images_to_combine

        images_data = fit_data_from_fit_images(raw_images[from_index:to_index])
        if len(crop_region) > 0:
            images_data = [crop(matrix, rows, cols) for matrix in images_data]

        # Calibrate images
        images_data = calibrateImages(
            masterDarkData=master_dark_data,
            masterFlatData=master_flat_data,
            listOfImagesData=images_data,
        )

        # Fill out the cropped regions with value of 1
        # Note, it's important to fill after the calibration step
        if len(crop_region) > 0:
            images_data = [fillMatrix(matrix, crop_region, 1) for matrix in images_data]

        # Alignment
        # We want to discard this set of images if any one image in this set cannot be aligned
        aligned_images_data = []
        for index, image_data in enumerate(images_data):
            try:
                aligned_images_data.append(imageAlignment(image_data, ref_image_path))
            except Exception as e:
                logging.error(f"Could not align image {raw_images[from_index + index]}")
                logging.error(f"Skipping combination {from_index}-{to_index}")
                break

        del images_data  # Delete unused object to free up memory

        # We proceed to next set of images if the alignment wasn't successful for any one
        # image in the combination set. We now this by checking no of aligned images.
        if len(aligned_images_data) < no_of_images_to_combine:
            continue

        # Combination
        combined_image_data = imageCombination(
            aligned_images_data,
            ALIGNED_COMBINED_OUTPUT_FOLDER
            / f"m23_7.0-{(to_index // no_of_images_to_combine):03}.fit",
            raw_images[from_index],  # Copy header from the first image in combination
        )
        logging.info(f"Combined images {from_index}-{to_index}")

        # Extraction
        extract_stars(
            combined_image_data,
            ref_file_path,
            save_as=LOG_FILES_COMBINED_OUTPUT_FOLDER
            / f"{night_date.strftime(LOG_FILE_COMBINED_FILENAME_DATE_FORMAT)}_m23_7.0-{(to_index // 10):03}.txt",
            radii_of_extraction=radii_of_extraction,
        )
        logging.info(f"Extraction from combination {from_index}-{to_index} completed")

    # Normalization
    normalization_helper(
        radii_of_extraction,
        FLUX_LOGS_COMBINED_OUTPUT_FOLDER,
        ref_file_path,
        [
            i
            for i in LOG_FILES_COMBINED_OUTPUT_FOLDER.glob("*")
            if i.is_file() and i.suffix == ".txt"
        ],
        night_date,
    )


def start_data_processing_auxiliary(config: Config):
    """
    This function processes (one or more) nights defined in config dict by
    putting together various functionalities like calibration, alignment,
    extraction, and normalization together.
    """

    OUTPUT_PATH: Path = config["output"]["path"]
    # If directory doesn't exist create directory including necessary parent directories.
    OUTPUT_PATH.mkdir(parents=True, exist_ok=True)

    for night in config["input"]["nights"]:
        night_path: Path = night["path"]
        night_date = get_date_from_input_night_folder_name(night_path.name)
        OUTPUT_NIGHT_FOLDER = OUTPUT_PATH / get_output_folder_name_from_night_date(night_date)
        # Create output folder for the night, if it doesn't already exist
        OUTPUT_NIGHT_FOLDER.mkdir(exist_ok=True)
        process_night(night, config, OUTPUT_NIGHT_FOLDER, night_date)


def start_data_processing(file_path: str):
    """
    Starts data processing with the configuration file `file_path` provided as the argument.
    Calls auxiliary function `start_data_processing_auxiliary` if the configuration is valid.
    """
    validate_file(Path(file_path), on_success=start_data_processing_auxiliary)



import numpy as np

from m23.trans import createFitFileWithSameHeader
from m23.utils import customMedian, fitDataFromFitImages

### This code is a direct implementation of steps
### mentioned in Handbook of Astronomical Image `Processing by
### Richard Berry and James Burnell version 2.0 section 6.3 Standard Calibration

###  makeMasterBias
###
###  purpose: creates masterBias, saves to fileName + returns masterBiasData
###
def makeMasterBias(saveAs, headerToCopyFromName=None, listOfBiasNames=None, listOfBiasData=None):

    if listOfBiasNames:
        listOfBiasData = fitDataFromFitImages(listOfBiasNames)

    if not listOfBiasNames and not listOfBiasData:
        raise Exception("Neither Bias data nor names were provided")

    if not headerToCopyFromName and listOfBiasNames:
        headerToCopyFromName = listOfBiasNames[0]
    elif not headerToCopyFromName and not listOfBiasNames:
        raise Exception("Filename to copy header from not provided")

    masterBiasData = getMedianOfMatrices(listOfBiasData)
    # headerToCopyFromName is the file whose header we're copying to
    #  save in masterBias
    createFitFileWithSameHeader(masterBiasData, saveAs, headerToCopyFromName)

    return masterBiasData


###  makeMasterDark
###
###  purpose: creates masterDark, saves to fileName + returns masterDarkData
###
### we generate the masterDark by "taking median of the dark frames"
###   --Richard Berry, James Burnell
def makeMasterDark(saveAs, headerToCopyFromName=None, listOfDarkNames=None, listOfDarkData=None):

    if listOfDarkNames:
        listOfDarkData = fitDataFromFitImages(listOfDarkNames)

    if not listOfDarkNames and not listOfDarkData:
        raise Exception("Neither Dark data nor names were provided")

    if not headerToCopyFromName and listOfDarkNames:
        headerToCopyFromName = listOfDarkNames[0]
    elif not headerToCopyFromName and not listOfDarkNames:
        raise Exception("Filename to copy header from not provided")

    masterDarkData = getMedianOfMatrices(listOfDarkData)
    # listOfDarks[0] is the file whose header we're copying to
    #  save in masterDark
    createFitFileWithSameHeader(masterDarkData, saveAs, headerToCopyFromName)

    return masterDarkData


###  makeMasterFlat
###
###  purpose: creates masterFlat, saves to fileName + returns masterFlatData
###
### we generate the masterFlat by
###   taking the median of flats and subtracting the masterDarkData
###
def makeMasterFlat(
    saveAs,
    masterDarkData,
    headerToCopyFromName=None,
    listOfFlatNames=None,
    listOfFlatData=None,
):

    ### We're supposed to use flat dark for the master flat
    ### but we did not take any for the new camera, so we're
    ### using dark frames instead
    ### In other words: If we don't have flat dark, use dark frames
    ###

    if listOfFlatNames:
        listOfFlatData = fitDataFromFitImages(listOfFlatNames)

    if not listOfFlatNames and not listOfFlatData:
        raise Exception("Neither Flat data nor names were provided")

    if not headerToCopyFromName and listOfFlatNames:
        headerToCopyFromName = listOfFlatNames[0]
    elif not headerToCopyFromName and not listOfFlatNames:
        raise Exception("Filename to copy header from not provied")

    firstFlatMedian = customMedian(listOfFlatData[0])
    ### We scale all flats w.r.t. first flat image
    ###   like the current IDL code does
    ###   https://github.com/LutherAstrophysics/idl-files/blob/f3d10e770d4d268908438deb4cda2076f21f1b14/master_calibration_frame_makerNEWEST.pro#L199
    listOfFlatData = [
        flatData * firstFlatMedian / customMedian(flatData) for flatData in listOfFlatData
    ]

    ### the we take the median of the scaled flats
    combinedFlats = getMedianOfMatrices(listOfFlatData)
    masterFlatData = combinedFlats - masterDarkData

    ### convert flat data to matrix of ints
    masterFlatData = np.array(masterFlatData, dtype="int")

    # listOfFlats[0] is the file whose header we're copying to
    #  save in masterDark
    createFitFileWithSameHeader(masterFlatData, saveAs, headerToCopyFromName)
    return masterFlatData


def getMedianOfMatrices(listOfMatrices):

    ## https://stackoverflow.com/questions/18461623/average-values-in-two-numpy-arrays
    return customMedian(np.array(listOfMatrices), axis=0, out=np.empty_like(listOfMatrices[0]))



import re
from datetime import date
from pathlib import Path

import numpy as np
import numpy.typing as npt

from m23.constants import FLUX_LOG_COMBINED_FILENAME_DATE_FORMAT
from m23.file.index import is_string_float


class FluxLogCombinedFile:
    """
    This class is instantiated with the string representing
    file path for the Flux Log Combined file that you want to analyze

    This object could be useful for analyzing values like the attendance of a
    star on a particular night, it's mean and median values on the night, etc.
    """

    # Class attributes
    header_rows = 6  # Specifies the first x rows that don't contain header information
    file_name_re = re.compile("(\d{2}-\d{2}-\d{2})_m23_7.0-ref_revised_71_(\d{4})_flux.txt")

    def __init__(self, path: str | Path) -> None:
        if type(path) == str:
            path = Path(path)
        self.__path = path
        self.__data = None
        self.__valid_data = None
        self.__read_data = False
        self.__attendance = None

    @classmethod
    def generate_file_name(cls, night_date: date, star_no: int):
        """
        Returns the file name to use for a given star night for the given night date
        """
        return f"{night_date.strftime(FLUX_LOG_COMBINED_FILENAME_DATE_FORMAT)}_m23_7.0-ref_revised_71_{star_no:04}_flux.txt"

    def _validate_file(self):
        if not self.path().exists():
            raise FileNotFoundError(f"File not found {self.path()}")
        if not self.path().is_file():
            raise ValueError("Directory provided, expected file f{self.path()}")

    def _calculate_attendance(self) -> float:
        """
        Calculates and returns the attendance for the night based on `self.data`
        Note that attendance is a value between 0-1.

        Preconditions:
            The object should have valid `self.data`
        Assumptions:
            `self.data` contains all data point albeit empty for a start for the night
        """
        data_points = len(self.data())
        positive_value_data_points = len(self.valid_data())
        return positive_value_data_points / data_points

    def read_file_data(self):
        """
        Reads the file and sets the the data attribute and attendance attribute in the object
        """
        self._validate_file()
        with self.path().open() as fd:
            lines = [line.strip() for line in fd.readlines()]
            lines = lines[self.header_rows :]  # Skip the header rows
            self.__data = np.array(lines, dtype="float")  # Save data as numpy array
            self.__valid_data = np.array(
                [x for x in self.__data if is_string_float(x) and float(x) > 0], dtype="float"
            )
        self.__read_data = True  # Marks file as read
        self.__attendance = self._calculate_attendance()

    def is_valid_file_name(self):
        """
        Checks if the file name is valid as per the file naming conventions
        of m23 data processing library. It returns the regex match pattern
        if the file name is valid.
        """
        return self.file_name_re.match(self.path.name)

    def star_number(self) -> int | None:
        """
        Returns the star number associated to the filename if the file name is valid
        """
        if self.is_valid_file_name():
            # The second capture group contains the star number
            return int(self.file_name_re.match(self.path.name)[2])

    def is_file_format_valid(self):
        """
        Checks if the file format is valid
        """
        return True

    # Accessors

    def path(self) -> Path:
        return self.__path

    def data(self) -> None | npt.ArrayLike:
        """
        The data property returns either None or a numpy one dimensional array
        """
        return self.__data

    def valid_data(self) -> None | npt.ArrayLike:
        """
        Returns a sample of data for the star for the nights with only valid data points > 0 magnitudes
        """
        return self.__valid_data

    def attendance(self) -> float:
        """
        Returns the attendance % (between 0-1) for star corresponding to the flux log combined file for the night
        """
        self._validate_file()
        if not self.__read_data:
            self.read_file_data()
        return self.__attendance

    def median(self) -> float | np.nan:
        """
        Returns the median value for the star for the night
        Note that this is the median of only valid data points (> 0 magnitudes)
        This means that 0.00 values are automatically ignored.
        Note that if the night doesn't contain any valid data, this returns nan
        """
        self._validate_file()
        if not self.__read_data:
            self.read_file_data()
        return np.median(
            self.valid_data()
        )  # Note to use only the valid data points to calculate median

    def mean(self) -> float | np.nan:
        """
        Returns the mean value for the star for the night
        Note that this is the median of only valid data points (> 0 magnitudes)
        This means that 0.00 values are automatically ignored.
        Note that if the night doesn't contain any valid data, this returns nan
        """
        self._validate_file()
        if not self.__read_data:
            self.read_file_data()
        return np.mean(
            self.valid_data()
        )  # Note to use only the valid data points to calculate mean

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"FluxLogCombinedFile {self.path()}"



import math
from functools import cache

import numpy as np

from m23.file import getLinesWithNumbersFromFile
from m23.matrix import blockRegions

# There are three methods of photometry
# 1. Aperture photometry: This is the method we use
# for the old camera, we create radii of 3, 4, 5 pixels around a star,
# and calculate its flux value
# 2. Profile photometry: TODO: This method was deemed
# inappropriate since the old images did not have
# as many pixels, but the new one have a lot more,
# so this is a method we want to try
# 3. Annular photometry


def newStarCenters(imageData, oldStarCenters):
    def centerFinder(position):
        x, y = position

        colWghtSum = 0
        rowWghtSum = 0
        WghtSum = 0
        for col in range(-5, 6):
            for row in range(-5, 6):
                if math.ceil(math.sqrt((col**2) + (row**2))) <= 5:
                    WghtSum += imageData[round(y) + row][round(x) + col]
                    colWghtSum += imageData[round(y) + row][round(x) + col] * (x + col)
                    rowWghtSum += imageData[round(y) + row][round(x) + col] * (y + row)

        if WghtSum > 0:
            xWght = colWghtSum / WghtSum
            yWght = rowWghtSum / WghtSum
        else:
            xWght = x
            yWght = y

        return yWght, xWght

    return [centerFinder(position) for position in oldStarCenters]


def extract_stars(
    image_data, reference_log_file_name, save_as, radii_of_extraction, image_name=None
):
    stars_positions_in_ref_file = starsPositionsInLogFile(reference_log_file_name)
    stars_centers_in_new_image = newStarCenters(image_data, stars_positions_in_ref_file)

    star_fluxes = {
        radius: flux_log_for_radius(radius, stars_centers_in_new_image, image_data)
        for radius in radii_of_extraction
    }
    no_of_stars = len(star_fluxes[radii_of_extraction[0]])

    # Create file
    with open(save_as, "w") as fd:
        # These 9 lines are to make sure it's the same format
        # as the old log files generated by IDL code
        fd.write("\n")
        fd.write(f"Star Data Extractor Tool: (Note: This program mocks format of AIP_4_WIN) \n")
        fd.write(f"\tImage {image_name or ''}:\n")
        fd.write(f"\tTotal no of stars: {no_of_stars}\n")
        fd.write(f"\tRadius of star diaphragm: {', '.join(map(str, radii_of_extraction))}\n")
        fd.write(f"\tSky annulus inner radius: \n")
        fd.write(f"\tSky annulus outer radius: \n")
        fd.write(f"\tThreshold factor: \n")

        headers = [
            "X",
            "Y",
            "XFWHM",
            "YFWHM",
            "Avg FWHM",
            "Sky ADU",
        ] + [f"Star ADU {radius}" for radius in radii_of_extraction]
        fd.write(
            f"{headers[0]:>16s}{headers[1]:>16s}{headers[2]:>16s}{headers[3]:>16s}{headers[4]:>16s}{headers[5]:>16s}{headers[6]:>16s}{headers[7]:>16s}{headers[8]:>16s}"
        )
        fd.write("\n")
        for star_index in range(no_of_stars):
            # We keep all stars in this step, filtering bad ones only in the normalization step
            data = (
                stars_centers_in_new_image[star_index][::-1]
                + (0, 0, 0)
                + (
                    star_fluxes[radii_of_extraction[0]][star_index][1],  # Sky ADU
                    star_fluxes[radii_of_extraction[0]][star_index][
                        2
                    ],  # Star ADU for first radius of extraction
                )
                + tuple(
                    [star_fluxes[i][star_index][2] for i in radii_of_extraction[1:]]
                )  # Star ADU for rest of the radii of extractions
            )
            fd.write(
                f"{data[0]:>16.2f}{data[1]:>16.2f}{data[2]:>16.4f}{data[3]:>16.4f}{data[4]:>16.4f}{data[5]:>16.2f}"
            )
            for star_adu in data[6:]:
                fd.write(f"{star_adu:>16.2f}")
            fd.write("\n")


def flux_log_for_radius(radius: int, stars_center_in_new_image, image_data):
    """
    TODO
    We need to optimize this code to work more efficiently with the caller function i.e extract_stars
    """

    regionSize = 64
    pixelsPerStar = np.count_nonzero(circleMatrix(radius))

    ###
    # return
    # an array of arrays of 64x64 matrices
    @cache
    def backgroundRegion():
        row, col = image_data.shape
        # block in third row first column can be accessed by [2, 0]
        return blockRegions(image_data, (regionSize, regionSize)).reshape(
            row // regionSize, col // regionSize, regionSize, regionSize
        )

    ###
    # backgroundRegionTuple is (2, 0) if referring to region in
    # third row, first column
    @cache
    def backgroundAverage(backgroundRegionTuple):
        row, column = backgroundRegionTuple
        region = backgroundRegion()[row][column]
        # throw out the background of zeroes, since
        # they might be at the edge
        sortedData = np.sort(region, axis=None)
        nonZeroIndices = np.nonzero(sortedData)
        # ignore the zeros
        sortedData = sortedData[nonZeroIndices]

        centeredArray = sortedData[
            int(len(sortedData) // 2 - 0.05 * len(sortedData)) : int(
                len(sortedData) // 2 + 0.05 * len(sortedData)
            )
        ]
        return np.mean(centeredArray)

    # starFlux
    #
    # return a three tuple
    # Total star flux + star pixels
    #   average background value for the star region
    #   star flux value after background subtraction
    def fluxSumForStar(position, radius):
        x, y = position
        starBox = image_data[x - radius : x + radius + 1, y - radius : y + radius + 1]
        starBox = np.multiply(starBox, circleMatrix(radius))
        backgroundAverageInStarRegion = backgroundAverage((x // regionSize, y // regionSize))
        subtractedStarFlux = np.sum(starBox) - backgroundAverageInStarRegion * pixelsPerStar
        ### Convert to zero, in case there's any nan... this ensures that
        ### two log files correspond to same star number as they are
        ### or after reading with something like getLinesWithNumbersFromFile
        ###
        ### This step makes our normalization code fast!
        return (
            np.nan_to_num(np.sum(starBox)),
            np.nan_to_num(backgroundAverageInStarRegion),
            np.nan_to_num(subtractedStarFlux),
        )

    stars_fluxes = [
        fluxSumForStar(np.round(position).astype("int"), radius)
        for position in stars_center_in_new_image
    ]
    return stars_fluxes


@cache
def starsPositionsInLogFile(fileName):
    linesWithNumbers = getLinesWithNumbersFromFile(fileName)
    # Assumes X and Y are the first two columns for the file
    ### NP WARNING
    ### IF dtype='float16' is specified, a number like '745.54' becomes 745.5
    ### Which means it gets rounded down, which is a HUGE problem if you're trying
    ### to match the output to IDL code
    ### So, we're using dtype='float'!!!
    return [np.array(line.split()[:2], dtype="float") for line in linesWithNumbers]


@cache
def circleMatrix(radius):
    lengthOfSquare = radius * 2 + 1
    myMatrix = np.zeros(lengthOfSquare * lengthOfSquare).reshape(lengthOfSquare, lengthOfSquare)
    for row in range(-radius, radius + 1):
        for col in range(-radius, radius + 1):
            if math.ceil(math.sqrt((row) ** 2 + (col) ** 2)) <= radius:
                myMatrix[row + radius][col + radius] = 1
    return myMatrix



import sys
from functools import cache
from pathlib import Path
from typing import Callable, List, TypedDict

import toml
from typing_extensions import NotRequired

from m23.constants import LOG_FILES_COMBINED_FOLDER_NAME
from m23.processor.config_loader import (
    is_night_name_valid,
    is_valid_radii_of_extraction,
)
from m23.utils import get_image_number_in_log_file_combined_file


class RenormalizeConfigProcessing(TypedDict):
    radii_of_extraction: List[int]


class RenormalizeConfigReference(TypedDict):
    file: Path | str


class RenormalizeConfigNight(TypedDict):
    path: Path | str
    first_logfile_number: int
    last_logfile_number: int
    files_to_use: NotRequired[List[Path]]


class RenormalizeConfigInput(TypedDict):
    nights: List[RenormalizeConfigNight]


class RenormalizeConfig(TypedDict):
    processing: RenormalizeConfigProcessing
    reference: RenormalizeConfigReference
    input: RenormalizeConfigInput


@cache
def get_relevant_log_files_combined_files(folder: Path, start: int, end: int) -> List[Path]:
    """
    Returns the list of log files combined files in the range enclosed
    (inclusively) by start and end path.

    param: folder: Path containing log files combined
    param: start: First file number to consider
    param: end: Last file number to consider

    Note that if start and/or end number fall of the edges of the available file
    numbers, the result contains the result that goes to the edge on the
    respective side.
    """
    result = []
    for file in folder.glob("*"):
        if file.is_file() and file.suffix == ".txt":
            if start <= get_image_number_in_log_file_combined_file(file) <= end:
                result.append(file)
    return result


def is_valid(config: RenormalizeConfig) -> bool:
    """
    Returns whether any error can be found in renormalize config dict
    """
    # Validate radii of extraction
    if not is_valid_radii_of_extraction(config["processing"]["radii_of_extraction"]):
        return False

    # Validate reference file
    ref_file = Path(config["reference"]["file"])
    if not (ref_file.exists() and ref_file.is_file() and ref_file.suffix == ".txt"):
        sys.stderr.write("Make sure the provided reference file exits and has txt extension\n")
        return False

    # Validate each night
    for night in config["input"]["nights"]:
        path = Path(night["path"])

        first_logfile = night["first_logfile_number"]
        last_logfile = night["last_logfile_number"]

        # Validate path
        if not path.exists():
            sys.stderr.write(f"Path {path} doesn't exist\n")
            return False

        # Check if the naming convention is valid for the input night
        if not is_night_name_valid(path):
            return False

        LOG_FILES_COMBINED_FOLDER = path / LOG_FILES_COMBINED_FOLDER_NAME

        if not LOG_FILES_COMBINED_FOLDER.exists():
            sys.stderr.write(f"Path {LOG_FILES_COMBINED_FOLDER} doesn't exist\n")
            return False

        if (type(first_logfile) != int or type(last_logfile) != int) and not (
            last_logfile >= first_logfile >= 1
        ):
            sys.stderr.write(
                f"Invalid logfile numbers. First: {first_logfile} Last: {last_logfile} for night {night}\n"
            )
            return False

        # Ensure that the list of files to use from LOG_FILES_COMBINED_FOLDER isn't empty
        # based on given constraints of start and end indices
        if (
            len(
                get_relevant_log_files_combined_files(
                    LOG_FILES_COMBINED_FOLDER, first_logfile, last_logfile
                )
            )
            == 0
        ):
            sys.stderr.write(f"No logfiles in range {first_logfile}-{last_logfile} \n")
            return False

        return True  # No errors detected


def sanity_check(config: RenormalizeConfig):
    """
    Performs any sanity check in the renormalization config
    """
    return config


def create_enhanced_config(config: RenormalizeConfig):
    """
    Creates an enhanced version of the renormalize config by converting str types to Path objects
    and other possible sanitation/optimizations
    """
    if type(config["reference"]["file"]) == str:
        config["reference"]["file"] = Path(config["reference"]["file"])

    for night in config["input"]["nights"]:
        if type(night["path"]) == str:
            night["path"] = Path(night["path"])
        # Set the list of log files to use for normalization based on the provided start, end indices
        night["files_to_use"] = get_relevant_log_files_combined_files(
            night["path"] / LOG_FILES_COMBINED_FOLDER_NAME,
            night["first_logfile_number"],
            night["last_logfile_number"],
        )

    return config


def validate_renormalize_config_file(
    file_path: Path, on_success: Callable[[RenormalizeConfig], None]
) -> None:
    """
    This method reads data processing configuration from the file path
    provided and calls the unary function on_success if the configuration
    file is valid with the configuration dictionary (Note, *not* config file).
    """
    if not file_path.exists() or not file_path.exists():
        raise FileNotFoundError("Cannot find configuration file")
    match toml.load(file_path):
        case {
            "processing": {"radii_of_extraction": list(_)},
            "input": {"nights": list(_)},
            "reference": {"file": str(_)},
        } as renormalize_config if is_valid(renormalize_config):
            on_success(sanity_check(create_enhanced_config(renormalize_config)))
        case _:
            sys.stderr.write("Stopping because the provided configuration file has issues.\n")


if __name__ == "__main__":
    validate_renormalize_config_file(Path("1.toml"), on_success=lambda *args: print("Success"))



from datetime import date

ASSUMED_MAX_BRIGHTNESS = 65_000

# Input folder/file name conventions
INPUT_CALIBRATION_FOLDER_NAME = "Calibration Frames"
M23_RAW_IMAGES_FOLDER_NAME = "m23"

# Date related settings
INPUT_NIGHT_FOLDER_NAME_DATE_FORMAT = "%B %d, %Y"
OUTPUT_NIGHT_FOLDER_NAME_DATE_FORMAT = "%B %d, %Y"
LOG_FILE_COMBINED_FILENAME_DATE_FORMAT = "%m-%d-%y"
FLUX_LOG_COMBINED_FILENAME_DATE_FORMAT = "%m-%d-%y"

# Output folder/file name conventions
CONFIG_FILE_NAME = "config.toml"
OUTPUT_CALIBRATION_FOLDER_NAME = "Calibration Frames"
ALIGNED_COMBINED_FOLDER_NAME = "Aligned Combined"
LOG_FILES_COMBINED_FOLDER_NAME = "Log Files Combined"
FLUX_LOGS_COMBINED_FOLDER_NAME = "Flux Logs Combined"
MASTER_DARK_NAME = "masterdark.fit"
MASTER_FLAT_NAME = "masterflat.fit"

# MISC
CAMERA_CHANGE_2022_DATE = date(2022, 6, 16)
TYPICAL_NEW_CAMERA_CROP_REGION = [
    [[0, 448], [0, 0], [492, 0], [210, 181]],
    [[0, 1600], [0, 2048], [480, 2048], [210, 1867]],
    [[1400, 2048], [2048, 2048], [2048, 1500], [1834, 1830]],
    [[1508, 0], [1852, 241], [2048, 521], [2048, 0]],
]



import astroalign as ast
import numpy as np
from astropy.io.fits import getdata as getfitsdata

### imageAlignment
###   purpose: align a particular image to a reference image
###
###   parameters:
###   imageToAlign: the image you want to align
###   refImage: the reference image (default is ref_revised_71)
###
###   returns
###   aligned Image data as fit file


def imageAlignment(imageToAlignData, refImageName):
    refImageData = getfitsdata(refImageName)

    ### workaround for endian type mismatch error in astroalign
    ### f4 means we're converting data to float 32
    target_fixed = np.array(refImageData, dtype="float")
    source_fixed = np.array(imageToAlignData, dtype="float")

    # alignedImageData, footprint = ast.register(imageToAlignData, refImageData, fill_value=50000)
    alignedImageData, footprint = ast.register(source_fixed, target_fixed, fill_value=0)

    return alignedImageData



import cv2 as cv
import numpy as np

### fillMatrix
###
### this is a wrapper our open cv's fillPoly function
###
### params:
###   matrix: (2d) matrix to work with
###   polygons: array of array of tuples (x, y) defining a polygon fill region
###   fillValue: value to fill
###
### example:
###   matrix:  np.arange(50).reshape(2, 5, 5)[1]
###   pol1: [(0,0), (1,1), (0,4)]
###   pol2: [(4, 0), (3, 1), (4,4)]
###   polygons: [pol1, pol2]
###   fillValue: 0
###
### returns:
###    Mutated original matrix


def fillMatrix(matrix, polygons, fillValue) -> None:
    return cv.fillPoly(matrix, np.array(polygons), fillValue)



import math
import os
from datetime import date
from pathlib import Path

import numpy as np

from m23.file import getLinesWithNumbersFromFile


###
### Function: normalizeLogFiles
###
### Takes in reference file name, log files names to normalize
### and folder path to save the new files in
###
### Produces a normalization.txt file and Flux Logs combined folder
### with normalized files in the saveFolder
###
### BIG ASSUMPTION!!!!
### We assume that noOfStars in all log files is the same as no of stars in
### reference file. This is the case if the log files were produced using extraction
### module in this library. However, in case of log files produced by previous IDL code
### this was't true. This is also one of the reasons, this normalization code might be
### faster than the IDL version.
###
###
### TODO: Mask out stars with center pixel not matching + crop the outlier stars
###  using linfit
def normalizeLogFiles(
    referenceFileName,
    logFilesNamesToNormalize,
    logfile_adu_column,
    saveFolder: Path,
    date_of_night: date,
    startImageUsed="",
    endImageUsed="",
):
    """
    This function normalizes the images provided.
    Note that the normalization isn't done with respect to the data in the reference image
    but with respect to some sample images take throughout the night.
    """
    # Wrapper around the function so we don't
    # have to keep passing the saveFolder as argument
    saveFileInFolder = lambda *args, **kwargs: saveFileInFolder(saveFolder, *args, **kwargs)

    # COLUMN_NUMBERS (0 means first column)
    ref_positions_xy = (0, 1)
    logfile_positions_xy = (0, 1)

    def aduInLogData(logData):
        return [float(line.split()[logfile_adu_column]) for line in logData]

    def starPositionsInLogData(logData):
        return [
            np.array(
                (line.split()[logfile_positions_xy[0] : logfile_positions_xy[1] + 1]),
                dtype="float",
            )
            for line in logData
        ]

    referenceData = getLinesWithNumbersFromFile(referenceFileName)
    logFilesData = [
        getLinesWithNumbersFromFile(logFileName) for logFileName in logFilesNamesToNormalize
    ]

    # At this point, if we want to look at 300-th star in our first image
    # we would do logFilesData[0][299] (beware of python index starting from 0)
    noOfFiles = len(logFilesData)

    # Normalization is done with reference to images 20%, 40%, 60% and 80% through night
    # The indices here are the index of the images from the night to which to normalize.
    # Note, we aren't normalizing with reference to the ref file
    indicesToNormalizeTo = np.linspace(0, noOfFiles, 6, dtype="int")[1:-1]
    adus_in_data_to_normalize_to = np.array(logFilesData, dtype="object")[indicesToNormalizeTo]

    # Get the ADU column in the logfiles
    adus_in_data_to_normalize_to = np.array(
        [aduInLogData(data) for data in adus_in_data_to_normalize_to]
    )

    referenceFileStarPositions = [
        np.array(line.split()[ref_positions_xy[0] : ref_positions_xy[1] + 1], dtype="float")
        for line in referenceData
    ]

    # Matrix of image by star so 4th star in 100th image will be normalized_star_data[99][3]
    normalized_star_data = np.zeros((noOfFiles, len(referenceData)))

    # Find normalization factor for each file
    allNormFactors = []
    for file_index in range(noOfFiles):
        # Normalization factor is the median of the scaleFactors of all stars for scaleFactors < 5
        # where scaleFactor for a star for that image is the ratio of that star's adu in
        # sum of data_to_normalize_to / 4 * adu in current image

        adu_of_current_log_file = np.array(aduInLogData(logFilesData[file_index]), dtype="float")

        starPositions = starPositionsInLogData(logFilesData[file_index])

        # Mask out stars with center more than 1 pixel away from those in the ref file
        # also mask if the star is outside the 12px box around the image
        for star_index in range(len(adu_of_current_log_file)):
            starX, starY = starPositions[star_index]
            refX, refY = referenceFileStarPositions[star_index]
            if math.sqrt((refX - starX) ** 2 + (refY - starY) ** 2) > 1:
                adu_of_current_log_file[star_index] = 0

        # Find normalization factor
        scale_factors_for_stars = np.sum(adus_in_data_to_normalize_to, axis=0) / (
            4 * adu_of_current_log_file
        )
        # Only get the median value for scale factors between 0 and 5, since some values are -inf or nan
        # We get the upper threshold 5 from the IDL code
        good_scale_factors = scale_factors_for_stars[
            np.where((scale_factors_for_stars < 5) & (scale_factors_for_stars > 0))
        ]
        normFactor = np.median(good_scale_factors) if len(good_scale_factors) else 0

        allNormFactors.append(normFactor)
        normalized_star_data[file_index] = normFactor * np.array(
            aduInLogData(logFilesData[file_index])
        )

    # Save the norm factor dot txt
    np.savetxt(
        saveFolder
        / f"{date_of_night.strftime('%m-%d-%y')}_m23_7.0-ref_revised_71_normfactors.txt",
        np.array(allNormFactors),
        fmt="%3.5f",
    )

    # Save the normalized data for each star
    noOfStars = len(normalized_star_data[0])
    for star_index in range(noOfStars):
        star_data = [
            normalized_star_data[file_index][star_index] for file_index in range(noOfFiles)
        ]
        # Turn all star_data that's negative to 0
        star_data = [currentData if currentData > 0 else 0 for currentData in star_data]

        # Add header information to the file
        # To make the layout consistent with output from IDL version
        with open(
            os.path.join(
                saveFolder,
                f"{date_of_night.strftime('%m-%d-%y')}_m23_7.0-ref_revised_71_{(star_index+1):04}_flux.txt",
            ),
            "w",
        ) as f:
            f.write(f"Program:\n")
            f.write(f"Started with image\t{startImageUsed}\n")
            f.write(f"Ended with image\t{endImageUsed}\n")
            f.write(f"Aligned with image {Path(referenceFileName).name}\n")
            f.write(f"X location:\t{''}\n")
            f.write(f"Y location:\t{''}\n")
            np.savetxt(
                f,
                np.array(star_data),
                fmt="%10.2f",
            )



import numpy as np

### imports from m23
from m23.trans import createFitFileWithSameHeader

### imagesToCombine
###   purpose: combine the calibrated images in a night
###
###   NOTE: The IDL code does the alignment before the
###   combination, so we will do the same
###
### paramaters
###   calibratedImageNames: file names of the calibrated images
###   imagesToCombineNumber: number of images to combine (default is 10)
###   totalImageNumber: total number of images to combine
###
### returns
###   the combined image data


def imageCombination(imagesData, fileName, fitFileNameToCopyHeaderFrom):
    imagesData = np.array(imagesData)
    combinedImageData = np.sum(imagesData, axis=0)

    createFitFileWithSameHeader(
        combinedImageData.astype("int"), fileName, fitFileNameToCopyHeaderFrom
    )
    return combinedImageData



import re
from pathlib import WindowsPath


def getLinesWithNumbersFromFile(fileName):
    with open(fileName, "r") as fd:
        allLines = [line.strip() for line in fd.readlines()]
        return list(
            filter(
                lambda line: re.search("\d+", line) and not re.search("[a-zA-Z]", line),
                allLines,
            )
        )


def formatWindowsPath(p: WindowsPath):
    # The anchor trickery was need to remove windows backward slash
    return "/".join([p.drive] + list(p.parts[1:]))


def is_string_float(string: str) -> bool:
    """
    Checks is the string is a float value.

    Example:
    is_string_float('1.1') -> True
    is_string_float('10a') -> False
    is_string_float('10') -> True
    is_string_float('1.1 .  ') -> True
    is_string_float('    1.7    ') -> True
    """
    try:
        float(string)
        return True
    except:
        return False



import os
import sys
from datetime import date
from functools import reduce
from pathlib import Path
from typing import Callable, Dict, List, TypedDict

import toml
from typing_extensions import NotRequired

from m23.constants import (
    CAMERA_CHANGE_2022_DATE,
    INPUT_CALIBRATION_FOLDER_NAME,
    M23_RAW_IMAGES_FOLDER_NAME,
    TYPICAL_NEW_CAMERA_CROP_REGION,
)
from m23.utils import (
    get_darks,
    get_date_from_input_night_folder_name,
    get_flats,
    get_raw_images,
)


# TYPE related to Config object described by the configuration file
class ConfigImage(TypedDict):
    rows: int
    columns: int
    crop_region: NotRequired[List[List[List[int]]]]


class ConfigProcessing(TypedDict):
    no_of_images_to_combine: int
    radii_of_extraction: List[int]


class ConfigInputNight(TypedDict):
    path: str | Path
    masterflat: NotRequired[str]


class ConfigInput(TypedDict):
    nights: List[ConfigInputNight]
    radii_of_extraction: List[int]


class ConfigReference(TypedDict):
    image: str | Path
    file: str | Path


class ConfigOutput(TypedDict):
    path: str | Path


class Config(TypedDict):
    image: ConfigImage
    processing: ConfigProcessing
    reference: ConfigReference
    input: ConfigInput
    output: ConfigOutput


def is_valid_radii_of_extraction(lst):
    """Verifies that each radius of extraction is a positive integer"""
    is_valid = all([type(i) == int and i > 0 for i in lst])
    if not is_valid:
        sys.stderr.write.write("Radius of extraction must be positive integers\n")
    return is_valid


def create_processing_config(config_dict: Config) -> Config:
    """
    Mutates `config_dict` to :
    1. Provide default values if optional values aren't provided.
    2. Replace all path str with Path objects
    """
    # Add empty list as the crop region if not present
    if not config_dict["image"].get("crop_region"):
        config_dict["image"]["crop_region"] = []

    # Convert input night str to Path objects
    for night in config_dict["input"]["nights"]:
        if type(night["path"]) == str:
            night["path"] = Path(night["path"])
        if night.get("masterflat") and type(night["masterflat"] == str):
            night["masterflat"] = Path(night["masterflat"])

    # Convert output path to Path object
    if type(config_dict["output"]["path"]) == str:
        config_dict["output"]["path"] = Path(config_dict["output"]["path"])

    # Convert reference file/img to Path object
    if type(config_dict["reference"]["file"]) == str:
        config_dict["reference"]["file"] = Path(config_dict["reference"]["file"])
    if type(config_dict["reference"]["image"]) == str:
        config_dict["reference"]["image"] = Path(config_dict["reference"]["image"])

    return config_dict


def prompt_to_continue(msg: str):
    sys.stderr.write(msg + "\n")
    response = input("Do you want to continue (y/yes to continue): ")
    if response.upper() not in ["Y", "YES"]:
        os._exit(1)


def sanity_check_image(config: ConfigImage, night_date: date):
    """
    Checks for abnormal values in configuration images
    """
    rows, cols = config["rows"], config["columns"]
    crop_region = config.get("crop_region")
    old_camera = night_date < CAMERA_CHANGE_2022_DATE
    if old_camera:
        if rows != 1024:
            prompt_to_continue(f"Detected non 1024 image row value for old camera date")
        if cols != 1024:
            prompt_to_continue(f"Detected non 1024 image column value for old camera date")
        if crop_region and type(crop_region) == list and len(crop_region) > 0:
            prompt_to_continue(f"Detected use of crop region for old camera.")
    else:
        if rows != 2048:
            prompt_to_continue(f"Detected non 2048 image row value for new camera date")
        if cols != 2048:
            prompt_to_continue(f"Detected non 2048 image column value for new camera date")
        if (
            not crop_region
            or crop_region
            and type(crop_region) != list
            or type(crop_region) == list
            and len(crop_region) == 0
        ):
            prompt_to_continue(
                f"We typically use crop images from new camera, you don't seem to define it"
            )
        else:
            try:
                for crop_section_index, crop_section in enumerate(crop_region):
                    for section_coordinate_index, section_coordinate in enumerate(crop_section):
                        if (
                            section_coordinate
                            != TYPICAL_NEW_CAMERA_CROP_REGION[crop_section_index][
                                section_coordinate_index
                            ]
                        ):
                            prompt_to_continue(
                                f"Mismatch between default crop region used in new camera and config file."
                            )
                            return  # Ignore further checking if already made the user aware of inconsistency once

            except Exception as e:
                prompt_to_continue(
                    f"Error while checking crop region with standard crop region value. {e}"
                )


def sanity_check(config_dict: Config) -> Config:
    """
    This method performs any sanity checks on the configuration file.
    """
    # Ensure sane values for rows/cols, etc.
    for night in config_dict["input"]["nights"]:
        night_date = get_date_from_input_night_folder_name(night["path"])
        sanity_check_image(config_dict["image"], night_date)
    return config_dict


def verify_optional_image_options(options: Dict) -> bool:
    """
    Verifies that the optional image options are valid
    """
    if len(options.keys()) > 1:
        return False
    crop_region: List[List[int]] = options.get("crop_region", [])
    # Ensure that all values in crop_region are non-negative integers
    values: List[int] = reduce(lambda prev, curr: prev + curr, crop_region, [])
    is_valid = all([type(i) == i and i >= 0 for i in values])
    if not is_valid:
        sys.stderr.write.write("Crop region must be an array of array of integers >= 0\n")
    return is_valid


def is_night_name_valid(NIGHT_INPUT_PATH: Path):
    """
    Returns if the input night folder name follows naming conventions.
    Prints msg to stderr if invalid.
    """
    # Check if the name of input folder matches the convention
    try:
        get_date_from_input_night_folder_name(NIGHT_INPUT_PATH.name)
        return True
    except:
        sys.stderr.write(
            f"Night {NIGHT_INPUT_PATH} folder name doesn't match the naming convention\n"
        )
        return False


def validate_night(night: ConfigInputNight) -> bool:
    """
    Checks whether the input configuration provided for night is valid.
    We check whether the input folders follow the required conventions,
    whether the right files are present and more.
    """
    NIGHT_INPUT_PATH = Path(night["path"])
    # Check if the night input path exists
    if not NIGHT_INPUT_PATH.exists():
        sys.stderr.write(f"Images path for {night} doesn't exist\n")
        return False

    if not is_night_name_valid(NIGHT_INPUT_PATH):
        return False

    CALIBRATION_FOLDER_PATH = NIGHT_INPUT_PATH / INPUT_CALIBRATION_FOLDER_NAME
    # Check if Calibration Frames exists
    if not CALIBRATION_FOLDER_PATH.exists():
        sys.stderr.write(f"Path {CALIBRATION_FOLDER_PATH} doesn't exist\n")
        return False

    M23_FOLDER_PATH = NIGHT_INPUT_PATH / M23_RAW_IMAGES_FOLDER_NAME
    # Check if m23 folder exists
    if not M23_FOLDER_PATH.exists():
        sys.stderr.write(f"Path {M23_FOLDER_PATH} doesn't exist\n")
        return False

    # Check for flats
    # Either the masterflat should be provided or the night should contain its own flats.
    if night.get("masterflat"):
        if not Path(night["masterflat"]).exists():
            sys.stderr.write(f"Provided masterflat path for {night} doesn't exist.\n")
            return False
    # If masterflat isn't provided, the night should have flats to use
    elif len(list(get_flats(CALIBRATION_FOLDER_PATH))) == 0:
        sys.stderr.write(
            f"Night {night} doesn't contain flats in {CALIBRATION_FOLDER_PATH}. Provide masterflat path.\n"
        )
        return False

    # Check for darks
    if len(list(get_darks(CALIBRATION_FOLDER_PATH))) == 0:
        sys.stderr.write(
            f"Night {night} doesn't contain darks in {CALIBRATION_FOLDER_PATH}. Cannot continue without darks.\n"
        )
        return False

    # Check for raw images
    try:
        if len(list(get_raw_images(M23_FOLDER_PATH))) == 0:
            sys.stderr.write(f"Night {night} doesn't have raw images in {M23_FOLDER_PATH}.\n")
            return False
    except ValueError as e:
        sys.stderr.write(
            "Raw image in night {night} doesn't confirm to 'something-00x.fit' convention.\n"
        )
        raise e

    return True  # Assuming we did the best we could to catch errors


def validate_input_nights(list_of_nights: List[ConfigInputNight]) -> bool:
    """
    Returns True if input for all nights is valid, False otherwise.
    """
    return all([validate_night(night) for night in list_of_nights])


def validate_reference_files(reference_image: str, reference_file: str) -> bool:
    """
    Returns True if reference_image and reference_file paths exist
    """
    img_path = Path(reference_image)
    file_path = Path(reference_file)
    if not (img_path.exists() and img_path.is_file() and img_path.suffix == ".fit"):
        sys.stderr.write("Make sure that the reference exists and has .fit extension")
        return False
    if not (file_path.exists() and file_path.is_file() and file_path.suffix == ".txt"):
        sys.stderr.write("Make sure that the reference file exists and has .txt extension")
        return False
    return True


def validate_file(file_path: Path, on_success: Callable[[Config], None]) -> None:
    """
    This method reads data processing configuration from the file path
    provided and calls the unary function on_success if the configuration
    file is valid with the configuration dictionary (Note, *not* config file).
    """
    if not file_path.exists() or not file_path.exists():
        raise FileNotFoundError("Cannot find configuration file")
    match toml.load(file_path):
        case {
            "image": {"rows": int(_), "columns": int(_), **optional_image_options},
            "processing": {
                "no_of_images_to_combine": int(_),
                "radii_of_extraction": list(radii_of_extraction),
            },
            "reference": {
                "image": str(reference_image),
                "file": str(reference_file),
            },
            "input": {"nights": list(list_of_nights)},
            "output": {"path": str(_)},
        } if (
            verify_optional_image_options(optional_image_options)
            and is_valid_radii_of_extraction(radii_of_extraction)
            and validate_input_nights(list_of_nights)
            and validate_reference_files(reference_image, reference_file)
        ):
            on_success(sanity_check(create_processing_config(toml.load(file_path))))
        case _:
            sys.stderr.write("Stopping because the provided configuration file has issues.\n")



import argparse
import sys
from pathlib import Path

from m23 import start_data_processing
from m23.processor import generate_masterflat, renormalize


def process(args):
    """
    This is a subcommand that handles data processing for one or more nights
    based on the configuration file path provided
    """
    config_file: Path = args.config_file
    if not config_file.exists():
        sys.stdout.write("Provided file doesn't exist\n")
        return
    if not config_file.is_file():
        sys.stdout.write("Invalid configuration file provided\n")
        return
    start_data_processing(config_file.absolute())


def norm(args):
    """
    This is a subcommand that handles renormalization for one or more nights
    based on the configuration file path provided
    """
    config_file: Path = args.config_file
    if not config_file.exists():
        sys.stdout.write("Provided file doesn't exist\n")
        return
    if not config_file.is_file():
        sys.stdout.write("Invalid configuration file provided\n")
        return
    renormalize(config_file.absolute())


def mf(args):
    """
    This is a subcommand that handles generating masterflat for a night from
    the flat images taken for the night
    """
    config_file: Path = args.config_file
    if not config_file.exists():
        sys.stdout.write("Provided file doesn't exist\n")
        return
    if not config_file.is_file():
        sys.stdout.write("Invalid configuration file provided\n")
        return
    generate_masterflat(config_file.absolute())


parser = argparse.ArgumentParser(prog="M23 Data processor", epilog="Made in Rapti")
subparsers = parser.add_subparsers()

# We are dividing our command line function into subcommands
# The first subcommand is `process` denoting a full fledged data processing for night(s)
process_parser = subparsers.add_parser("process", help="Process raw data for one or more nights")
process_parser.add_argument(
    "config_file", type=Path, help="Path to toml configuration file for data processing"
)  # positional argument
# Adding a default value so we later know which subcommand was invoked
process_parser.set_defaults(func=process)

# Renormalize parser
norm_parser = subparsers.add_parser(
    "norm", help="Normalize log files combined for one or more nights"
)
norm_parser.add_argument(
    "config_file", type=Path, help="Path to toml configuration file for renormalization"
)  # positional argument
# Adding a default value so we later know which subcommand was invoked
norm_parser.set_defaults(func=norm)

# Masterflat generator parser
mf_parser = subparsers.add_parser("mf", help="Generate masterflat for a night from its raw flats")
mf_parser.add_argument(
    "config_file", type=Path, help="Path to toml configuration file for renormalization"
)  # positional argument
# Adding a default value so we later know which subcommand was invoked
mf_parser.set_defaults(func=mf)

args = parser.parse_args()
if hasattr(args, "func"):
    args.func(args)
else:
    parser.parse_args(["-h"])



import logging
from pathlib import Path

from m23.constants import FLUX_LOGS_COMBINED_FOLDER_NAME
from m23.processor.process_nights import normalization_helper
from m23.utils import get_date_from_input_night_folder_name, get_log_file_name

from .renormalize_config_loader import (
    RenormalizeConfig,
    validate_renormalize_config_file,
)


def renormalize_auxiliary(renormalize_dict: RenormalizeConfig):
    for night in renormalize_dict["input"]["nights"]:
        NIGHT_FOLDER = night["path"]
        night_date = get_date_from_input_night_folder_name(NIGHT_FOLDER.name)
        log_file_path = NIGHT_FOLDER / get_log_file_name(night_date)
        radii_of_extraction = renormalize_dict["processing"]["radii_of_extraction"]
        logging.basicConfig(
            filename=log_file_path,
            format="%(asctime)s %(message)s",
            level=logging.INFO,
        )
        logging.info(f"Running renormalization for radii {radii_of_extraction}")

        LOG_FILES_COMBINED_FOLDER: Path = NIGHT_FOLDER / FLUX_LOGS_COMBINED_FOLDER_NAME
        # Create log files combined folder if it doesn't yet exist
        LOG_FILES_COMBINED_FOLDER.mkdir(exist_ok=True)

        normalization_helper(
            radii_of_extraction,
            LOG_FILES_COMBINED_FOLDER,
            renormalize_dict["reference"]["file"],
            night["files_to_use"],
            night_date,
        )


def renormalize(file_path: str):
    """
    Starts renormalization with the configuration file `file_path` provided as the argument.
    Calls auxiliary function `renormalize_auxiliary` if the configuration is valid.
    """
    validate_renormalize_config_file(Path(file_path), on_success=renormalize_auxiliary)



import numpy as np


def crop(matrix, row, column):
    return matrix[:row, :column]


### cropIntoRectangle
###
### parameters:
###   matrix: a numpy array
###   x: xCordinate to start crop
###   xLength: length across first axis
###   y: yCordiate to start crop
###   yLength: length across second (y) axis
###
### returns
###   cropped rectangle
def cropIntoRectangle(matrix, x, xLength, y, yLength):
    return matrix[x : x + xLength, y : y + yLength]



import sys
from pathlib import Path
from typing import Callable, TypedDict

import toml

from m23.constants import INPUT_CALIBRATION_FOLDER_NAME
from m23.processor.config_loader import (
    ConfigImage,
    is_night_name_valid,
    sanity_check_image,
)
from m23.utils import get_date_from_input_night_folder_name, get_flats


class MasterflatGeneratorConfig(TypedDict):
    input: Path | str
    output: Path | str
    image: ConfigImage


def is_valid(config: MasterflatGeneratorConfig) -> bool:
    """
    Returns whether the configuration file is valid
    """
    NIGHT_INPUT_PATH = Path(config["input"])
    if not NIGHT_INPUT_PATH.exists():
        sys.stderr.write(f"Input folder {NIGHT_INPUT_PATH} doesn't exist.\n")
        return False

    # Verify that the Night folder name matches the naming convention
    if not is_night_name_valid(NIGHT_INPUT_PATH):
        # No error message needed as the function `is_night_name_valid` write error if there's one
        return False

    # Verify that the CALIBRATION FOLDER exists
    CALIBRATION_FOLDER_PATH = NIGHT_INPUT_PATH / INPUT_CALIBRATION_FOLDER_NAME
    if not CALIBRATION_FOLDER_PATH.exists():
        sys.stderr.write(f"Calibration folder {CALIBRATION_FOLDER_PATH} doesn't exist.\n")
        return False

    # Verify that the night contains flats to use
    if len(list(get_flats(CALIBRATION_FOLDER_PATH))) == 0:
        sys.stderr.write(
            f"Night {NIGHT_INPUT_PATH} doesn't contain flats in {CALIBRATION_FOLDER_PATH}. Provide masterflat path.\n"
        )
        return False

    try:
        output_path = Path(config["output"])
        output_path.mkdir(parents=True, exist_ok=True)  # Create directory if not exists
    except Exception as e:
        sys.stderr.write(f"Error in output folder. {e} \n")
        return False

    if not is_image_properties_valid(config["image"]):
        # No error message needed as the function `is_image_properties_valid` write error if there's one
        return False

    return True  # No errors detected


def is_image_properties_valid(image_config: ConfigImage) -> bool:
    """
    Checks and returns if  the image_properties is valid.
    If invalid, write the error msg in stderr.
    """

    # Validate the image properties in the configuration file
    # Ensure that rows and cols are int > 0
    rows, cols = image_config["rows"], image_config["columns"]
    if type(rows) != int or type(cols) != int or rows <= 0 or cols <= 0:
        sys.stderr.write(
            f"Rows and columns of image have to be > 0. Got  rows:{rows} cols:{cols}\n"
        )
        return False
    # Ensure that if crop_region is present, it has to be list of list of list of ints
    if crop_region := image_config.get("crop_region"):
        try:
            for i in crop_region:
                for j in i:
                    valid_values = all([type(x) == int and x >= 0 for x in j])
                    if not valid_values:
                        sys.stderr.write(f"Invalid value detected in crop_region {j}.\n")
                        return False
        except Exception as e:
            sys.stderr.write(f"Error in crop_region {j}.\n")
            return False

    return True  # No error detected


def create_enhanced_config(config: MasterflatGeneratorConfig) -> MasterflatGeneratorConfig:
    """
    This function enhances the configuration file for ease of functions
    that later require processing of the config file
    """
    # Covert folder str to path
    config["input"] = Path(config["input"])
    config["output"] = Path(config["output"])
    return config


def sanity_check(config: MasterflatGeneratorConfig) -> MasterflatGeneratorConfig:
    """
    This method is warn about technically correct but abnormal configuration values
    """
    night_date = get_date_from_input_night_folder_name(config["input"])
    sanity_check_image(config["image"], night_date)
    return config


def validate_generate_masterflat_config_file(
    file_path: Path, on_success: Callable[[MasterflatGeneratorConfig], None]
):
    """
    This method reads configuration file for generating masterflat
    and if the config is valid, calls the unary on_success function with the
    configuration file
    """

    if not file_path.exists() or not file_path.exists():
        raise FileNotFoundError("Cannot find configuration file")
    match toml.load(file_path):
        case {
            "input": _,
            "output": _,
            "image": {"rows": int(_), "columns": int(_), **optional_image_options},
        } as masterflat_generator_config if is_valid(masterflat_generator_config):
            on_success(sanity_check(create_enhanced_config(masterflat_generator_config)))
        case _:
            sys.stderr.write("Stopping because the provided configuration file has issues.\n")



from .calibration import calibrateImages
from .master_calibrate import makeMasterBias, makeMasterDark, makeMasterFlat

__all__ = ["makeMasterBias", "makeMasterDark", "makeMasterFlat", "calibrateImages"]



import os


### Dry specifies if you want to actually make the changes
### or just see what the changes will be
###
def rename(folder, newNameFromOldName, dry=True):
    files = os.listdir(folder)
    for file in files:
        if dry:
            print(f"{file} -> {newNameFromOldName(file)}")
        else:
            renameFile(file, folder, newNameFromOldName)


def renameFile(fileName, folderName, renameFunction):
    newName = renameFunction(fileName)
    oldFilePath = os.path.join(folderName, fileName)
    newFilePath = os.path.join(folderName, newName)
    os.rename(oldFilePath, newFilePath)
